//
//  TNRequestOperation.m
//  ThinkNear
//
//  Created by Adam Ernst on 3/7/11.
//  Copyright 2011 cosmicsoft. All rights reserved.
//

#import "TNRequestOperation.h"
#import "CJSONDeserializer.h"


NSString *TNRequestOperationErrorDomain = @"TNRequestOperationErrorDomain";


@implementation TNRequestOperation

- (NSError *)error {
    return error;
}

#pragma mark Abstract Methods

- (OAMutableURLRequest *)request {
    [NSException raise:NSInternalInconsistencyException 
                format:@"TNRequestOperation -request"];
    return nil;
}

- (void)handleResponse:(NSHTTPURLResponse *)response data:(id)jsonData {
    [NSException raise:NSInternalInconsistencyException 
                format:@"TNRequestOperation -handleResponse:"];
}

#pragma mark Helper Methods

- (NSString *)endpointURL {
    NSString *serverIp = [[NSUserDefaults standardUserDefaults] objectForKey:@"server_address"];
    if (serverIp && [serverIp length]) {
        return [NSString stringWithFormat:@"http://%@", serverIp];
    }
    return @"http://api-sandbox.thinknear.com";
}

- (OAConsumer *)consumer {
    NSDictionary *infoDict = [[NSBundle mainBundle] infoDictionary];
    return [[[OAConsumer alloc] initWithKey:[infoDict objectForKey:@"TNConsumerKey"]
                                     secret:[infoDict objectForKey:@"TNConsumerSecret"]] autorelease];
}

- (OAMutableURLRequest *)JSONRequestWithRelativeURL:(NSString *)url {
    assert([[url substringToIndex:1] isEqualToString:@"/"]);
    NSURL *fullURL = [NSURL URLWithString:[NSString stringWithFormat:@"%@/1.0%@", 
                                           [self endpointURL], url]];
    
    OAMutableURLRequest *r;
    r = [[[OAMutableURLRequest alloc] initWithURL:fullURL
                                         consumer:[self consumer] 
                                            token:nil 
                                            realm:nil
                                signatureProvider:nil] autorelease];
    [r setTimeoutInterval:30.0];
    [r addValue:@"application/json" forHTTPHeaderField:@"Accept"];
    return r;
}

#pragma mark NSOperation

// |TNRequestOperation| is concurrent, meaning you can call -start from the 
// main thread. (Detect completion with -setCompletionBlock, or observe
// -isFinished.)
- (BOOL)isConcurrent {
    return YES;
}

- (BOOL)isExecuting {
    return (connection != nil);
}

- (BOOL)isFinished {
    return finished;
}

- (void)finishWithError:(NSError *)err {
    assert(finished == NO);
    assert(error == nil);
    
    BOOL wasExecuting = (connection != nil);
    
    if (wasExecuting) [self willChangeValueForKey:@"isExecuting"];
    [self willChangeValueForKey:@"isFinished"];
    
    finished = YES;
    error = [err retain];
    if (wasExecuting) {
        [connection release], connection = nil;
        [response release], response = nil;
        [data release], data = nil;
    }
    
    [self didChangeValueForKey:@"isFinished"];
    if (wasExecuting) [self didChangeValueForKey:@"isExecuting"];
}

- (void)cancelConnection {
    assert([NSThread isMainThread]);
    
    if (connection) { // Equivalent to [self isExecuting]
        [connection cancel];
        [self finishWithError:[NSError errorWithDomain:NSCocoaErrorDomain
                                                  code:NSUserCancelledError
                                              userInfo:nil]];
    }
}

- (void)cancel {
    [self performSelectorOnMainThread:@selector(cancelConnection)
                           withObject:nil 
                        waitUntilDone:NO];
    
    [super cancel];
}

- (void)start {
    OAMutableURLRequest *request = [self request];
    assert(request != nil);
    assert(connection == nil);
    assert(finished == NO);
    
    if ([self isCancelled]) {
        [self finishWithError:[NSError errorWithDomain:NSCocoaErrorDomain 
                                                  code:NSUserCancelledError
                                              userInfo:nil]];
    } else {
        [self willChangeValueForKey:@"isExecuting"];
        [request prepare]; // Sign for OAuth
        assert((connection = [[NSURLConnection alloc] initWithRequest:request
                                                             delegate:self 
                                                     startImmediately:NO]));
        
        // Schedule in the main thread's run loop. If we're in an 
        // NSOperationQueue, our run loop is transient.
        [connection scheduleInRunLoop:[NSRunLoop mainRunLoop] 
                              forMode:NSDefaultRunLoopMode];
        [connection start];
        [self didChangeValueForKey:@"isExecuting"];
    }
}

#pragma mark NSURLConnection delegate

- (void)connection:(NSURLConnection *)conn didReceiveResponse:(NSURLResponse *)resp {
    assert([NSThread isMainThread]);
    assert(connection == conn);
    assert([resp isKindOfClass:[NSHTTPURLResponse class]]);
    
    [response release], response = [(NSHTTPURLResponse *) resp retain];
    
    if (data) {
        [data setLength:0];
    }
}

- (void)connection:(NSURLConnection *)conn didReceiveData:(NSData *)receivedData {
    assert([NSThread isMainThread]);
    assert(connection == conn);
    assert(receivedData != nil);
    
    if (!data) {
        data = [[NSMutableData alloc] init];
    }
    [data appendData:receivedData];
}

- (void)connectionDidFinishLoading:(NSURLConnection *)conn {
    assert([NSThread isMainThread]);
    assert(connection == conn);
    
    assert(response != nil);
    assert(data != nil);
    if ([response statusCode] == 401) {
        [self finishWithError:[NSError errorWithDomain:TNRequestOperationErrorDomain
                                                  code:TNRequestOperationErrorInvalidAuth
                                              userInfo:nil]];
    } else if ([response statusCode] / 100 != 2) {
        NSDictionary *userInfo = [NSDictionary dictionaryWithObject:[NSString stringWithFormat:@"Unexpected HTTP status code %d", [response statusCode]]
                                                             forKey:NSLocalizedDescriptionKey];
        
        [self finishWithError:[NSError errorWithDomain:TNRequestOperationErrorDomain 
                                                  code:TNRequestOperationErrorUnexpectedStatusCode
                                              userInfo:userInfo]];
    } else if (![[response MIMEType] isEqualToString:@"application/json"]) {
        [self finishWithError:[NSError errorWithDomain:TNRequestOperationErrorDomain 
                                                  code:TNRequestOperationErrorNotJson 
                                              userInfo:nil]];
    } else {
        NSError *theError = nil;
        id jsonResponse = [[CJSONDeserializer deserializer] deserialize:data error:&theError];
        
        if (jsonResponse) {
            [self handleResponse:response data:jsonResponse];
        } else {
            [self finishWithError:theError];
        }
    }
    assert(finished && connection == nil && response == nil && data == nil);
}

- (void)connection:(NSURLConnection *)conn didFailWithError:(NSError *)err {
    assert([NSThread isMainThread]);
    assert(connection == conn);
    assert(err != nil);
    
    [self finishWithError:err];
}

@end
